!---------------------------------------------------------------
!
!	The holographic exemplar model as presented at SCiP 2011:
!
!	Chubala, C. M., & Jamieson, R. K. (2011, November). The
!	holographic exemplar model: Recoding and representation in
!	artificial grammar learning. Poster delivered at the 41st
!	Annual Meeting of the Society for Computers in Psychology,
!	Seattle, Washington, USA.
!
!	Published in Behaviour Research Methods:
!	
!	Chubala, C. M., & Jamieson, R. K. (2012). Recoding and
!	representation in artificial grammar learning. Behavior
!	Resarch Methods 
!
!---------------------------------------------------------------
PROGRAM Holographic_exemplar_model
use number_generators
implicit none

  integer, parameter	::	N_features = 100,									&	
							N_length = 10,										&
							N_letters = 255,									&
							N_subjects = 100,									&
							N_orders = 10,										&
							Max_items = 100

  real					::	Memory(N_features,Max_items),						&
							Letter_representations(N_features,N_letters),		&
							Order_representations(N_features,N_orders),			&
							Sum_echo_intensities(Max_items),					&
							Echo_intensities(Max_items),						&
							Mean,												&
							SD,													&
							L
							
  integer				::	i,													&
							Training_strings(N_length,Max_items),				&
							Test_strings(N_length,Max_items),					&
							N_study,											&
							N_test, 											&
							P1(N_features), 									&
							P2(N_features),										&
							Encoding_strategy
							
  character(30)			::	Output_file_name, Training, Test

  	call RandSeed 
	
	Mean = 0.0
  	SD = 1.0/Sqrt(float(N_features))

	call Gather_IO
	
	Sum_echo_intensities = 0.0
	   	do i = 1, N_subjects
			call Make_scramble (P1, P2)			
			call Get_letter_representations
			call Get_order_representations
			call Store_training_items_to_memory
			call Get_echo_intensities (Echo_intensities)
			Sum_echo_intensities = Sum_echo_intensities + Echo_intensities
		enddo

	call Write_output_to_file 

!*********
Contains

!---------
subroutine Gather_IO
implicit none

	write(*,*)
	write(*,*)
	write(*,'(A)') 'File specification:'
	write(*,'(A)') '-------------------'
	write(*,*)
	write(*,'(A)',advance='no') 'Name the output file: '
	read(*,*) Output_file_name
	write(*,'(A)',advance='no') 'Name the training file: '
	read(*,*) Training
	call Process_stimulus_file(Training, Training_strings, N_study)
	write(*,'(A)',advance='no') 'Name the test file: '
	read(*,*) Test	
	call Process_stimulus_file(Test, Test_strings, N_test)

	write(*,*)
	write(*,*)
	write(*,'(A)') 'Encoding options: '
	write(*,'(A)') '-----------------'
	write(*,*)
		write(*,'(A)') '1 = a*b*1 + c*d*2 + e*f*3'
		write(*,'(A)') '2 = a*b*c*1 + d*e*f*2'
		write(*,'(A)') '3 = a*1 + b*2 + c*3 + d*4'
		write(*,'(A)') '4 = a*b*1 + c*d*2'
	write(*,*)	
	write(*,'(A)',advance='no') 'Enter encoding strategy: '
	read(*,*) Encoding_strategy

	write(*,*)	
	write(*,*)
	write(*,'(A)') 'Parameter specification:'
	write(*,'(A)') '------------------------'
	write(*,*)
	write(*,'(A)', advance='no') 'Learning parameter (L) = '
	read(*,*) L
	write(*,*)

return
end subroutine Gather_IO

!---------
subroutine Get_letter_representations
implicit none
	integer		::	i, j
	
	Letter_representations = 0.0
	do i = 1, N_letters
		do j = 1, N_features
			Letter_representations(j,i) = Gaussian(Mean, SD)
		enddo
	enddo

return
end subroutine Get_letter_representations

!---------
subroutine Get_order_representations
implicit none
	integer		::	i, j
	
	Order_representations = 0.0
	do i = 1, N_orders
		do j = 1, N_features
			Order_representations(j,i) = Gaussian(Mean, SD)
		enddo
	enddo

return
end subroutine Get_order_representations

!----------
subroutine Store_training_items_to_memory
implicit none
	integer		:: i, j
	
	Memory = 0.0
	do i = 1, N_study
		if (Encoding_strategy == 1) call Encode_AB_CD_EF (Memory(:,i), Training_strings(:,i))
		if (Encoding_strategy == 2) call Encode_ABC_DEF  (Memory(:,i), Training_strings(:,i))
		if (Encoding_strategy == 3) call Encode_A_B_C_D  (Memory(:,i), Training_strings(:,i))
		if (Encoding_strategy == 4) call Encode_AB_CD    (Memory(:,i), Training_strings(:,i))
	enddo
	
	do i = 1, N_study
		do j = 1, N_features
			if (flat(1.0) > L) Memory(j,i) = 0.0
		enddo
	enddo

return
end subroutine Store_training_items_to_memory

!----------
subroutine Get_echo_intensities (oVec)
implicit none
	integer		:: 	i, j
	real		:: 	Probe(N_features), Echo(N_features), Activation, oVec(N_test)
	
	do i = 1, N_test

		Probe = 0.0
		if (Encoding_strategy == 1) call Encode_AB_CD_EF (Probe, Test_strings(:,i))
		if (Encoding_strategy == 2) call Encode_ABC_DEF  (Probe, Test_strings(:,i))
		if (Encoding_strategy == 3) call Encode_A_B_C_D  (Probe, Test_strings(:,i))
		if (Encoding_strategy == 4) call Encode_AB_CD    (Probe, Test_strings(:,i))
				
		Echo = 0.0
		do j = 1, N_study
			Activation = Cosine(Probe(:), Memory(:,j), N_features)**3
			Echo(:) = Echo(:) + Activation * Memory(:,j)
		enddo
				
		oVec(i) = Cosine(Probe(:), Echo(:), N_features)
			
	enddo

return
end subroutine Get_echo_intensities

!----------
subroutine Write_output_to_file
implicit none
	
	Open(1, file=Output_file_name)    
		do i = 1, N_test			
			write(1,'(10A, 100F12.4)') Digits_to_letters(Test_strings(:,i)), Sum_echo_intensities(i)/N_subjects
		enddo
	Close(1)
	
	write(*,*)	
	write(*,*)
	write(*,'(A)') 'PROGRAM COMPLETED'
	write(*,*)
	write(*,*)
	

return
end subroutine Write_output_to_file

!---------
function Convolve (v1, v2, n)
implicit none
	integer 					:: i, j, n
  	real, dimension(0:n-1) 		:: Convolve, Summ, v1, v2
 
	Summ = 0.0

	do i = 0, (n-1)
		do j = 0, (n-1)
			Summ(i) = Summ(i) + v1(modulo(j,n)) * v2(modulo(i-j,n))	
		enddo
	enddo
 
	Convolve = Summ
 
return
end function Convolve

!--------
function Cosine (v1, v2, n)
implicit none
	integer		:: n
	real		:: v1(n), v2(n), x(n), y(n), Cosine
	
	Cosine = 0.0

	if (sum(abs(v1)) /= 0.0 .and. sum(abs(v2)) /= 0.0) then

	x = v1/sqrt(dot_product(v1,v1))
	y = v2/sqrt(dot_product(v2,v2))
	
	Cosine = Dot_product(x,y)
	
	endif
		
return
end function Cosine

!--------
function Scramble (oVec, iVec)
 implicit none
 	real, dimension(N_features)	:: oVec, Tmp, Scramble
 	integer						:: i, iVec(N_features)
 
 	do i = 1, N_features
 	  Tmp(i) = oVec(iVec(i))
 	enddo
 
 	Scramble = Tmp
 
return
end function Scramble

!--------
subroutine Make_scramble (v1, v2)
 implicit none
	integer		:: i, j, v1(N_features), v2(N_features), Raw_order(N_features)

	!--- For v1:
	do i = 1, N_features
		Raw_order(i) = i
	enddo
	
	do i = 1, N_features
		do 
			j = FlatInt(N_features)
			if (Raw_order(j) <> -1) exit
		enddo
	  Raw_order(j) = -1
	  v1(i) = j
	enddo
		
	!--- For v2:
	do i = 1, N_features
		Raw_order(i) = i
	enddo
	
	do i = 1, N_features
		do 
			j = FlatInt(N_features)
			if (Raw_order(j) <> -1) exit
		enddo
	  Raw_order(j) = -1
	  v2(i) = j
	enddo	
	
return
end subroutine Make_scramble

!--------
subroutine Process_stimulus_file (fName, oMat, nitems)
implicit none
 	integer			:: 	i, nitems, Tmp_int(N_length), oMat(N_length,Max_items)				
  	character		:: 	Tmp_chr(N_length)
	character(30)	::	fName
	
	oMat = 0
	Open(Unit=2, File=fName)
		read(Unit=2,Fmt='(I3)') nitems
			do i = 2, nitems+1		
				read(Unit=2,Fmt='(10A)') Tmp_chr
				oMat(:,i-1) = Letters_to_digits(Tmp_chr)				
			enddo
	Close(2)

return
end subroutine Process_stimulus_file

!--------
subroutine Encode_AB_CD_EF (oVec, iVec)
implicit none
	integer		:: iVec(N_length)
	real		:: oVec(N_features), Tmp(N_features)

	oVec(:) = 0.0

	Tmp = Convolve(Scramble(Letter_representations(:,iVec(1)), P1), &
							Scramble(Letter_representations(:,iVec(2)), P2), N_features)		
	oVec = oVec + Convolve(Tmp, Order_representations(:,1), N_features)

	Tmp = Convolve(Scramble(Letter_representations(:,iVec(3)), P1), &
							Scramble(Letter_representations(:,iVec(4)), P2), N_features)		
	oVec = oVec + Convolve(Tmp, Order_representations(:,2), N_features)
	
	Tmp = Convolve(Scramble(Letter_representations(:,iVec(5)), P1), &
							Scramble(Letter_representations(:,iVec(6)), P2), N_features)		
	oVec = oVec + Convolve(Tmp, Order_representations(:,3), N_features)

return
end subroutine Encode_AB_CD_EF

!--------
subroutine Encode_ABC_DEF (oVec, iVec)
implicit none
 integer	:: iVec(N_length)
 real		:: oVec(N_features), Tmp(N_features)

	oVec(:) = 0.0

	Tmp = Convolve(Scramble(Letter_representations(:,iVec(1)), P1), &
							Scramble(Letter_representations(:,iVec(2)), P2), N_features)		
	Tmp = Convolve(Scramble(Tmp, P1), Scramble(Letter_representations(:,3), P2), N_features)
	oVec = oVec + Convolve(Tmp, Order_representations(:,1), N_features)
	
	Tmp = Convolve(Scramble(Letter_representations(:,iVec(4)), P1), &
							Scramble(Letter_representations(:,iVec(5)), P2), N_features)
	Tmp = Convolve(Scramble(Tmp, P1), Scramble(Letter_representations(:,6), P2), N_features)
	oVec = oVec + Convolve(Tmp, Order_representations(:,2), N_features)

return
end subroutine Encode_ABC_DEF

!--------
subroutine Encode_A_B_C_D (oVec, iVec)
implicit none
	integer		:: iVec(N_length)
	real		:: oVec(N_features), Tmp(N_features)

	oVec(:) = 0.0

	oVec = oVec + Convolve(Letter_representations(:,iVec(1)), Order_representations(:,1), N_features)
	oVec = oVec + Convolve(Letter_representations(:,iVec(2)), Order_representations(:,2), N_features)
	oVec = oVec + Convolve(Letter_representations(:,iVec(3)), Order_representations(:,3), N_features)
	oVec = oVec + Convolve(Letter_representations(:,iVec(4)), Order_representations(:,4), N_features)

return
end subroutine Encode_A_B_C_D

!--------
subroutine Encode_AB_CD (oVec, iVec)
implicit none
 integer	:: iVec(N_length)
 real		:: oVec(N_features), Tmp(N_features)

	oVec(:) = 0.0

	Tmp = Convolve(Scramble(Letter_representations(:,iVec(1)), P1), &
							Scramble(Letter_representations(:,iVec(2)), P2), N_features)		
	oVec = oVec + Convolve(Tmp, Order_representations(:,1), N_features)

	Tmp = Convolve(Scramble(Letter_representations(:,iVec(3)), P1), &
							Scramble(Letter_representations(:,iVec(4)), P2), N_features)		
	oVec = oVec + Convolve(Tmp, Order_representations(:,2), N_features)

return
end subroutine Encode_AB_CD

!-----
function Letters_to_digits (iVec)
implicit none
 	integer		:: i, Letters_to_digits(N_length)
 	character	:: iVec(N_length)

	do i = 1, N_length
		Letters_to_digits(i) = ichar(iVec(i))
	enddo	

return
end function Letters_to_digits

!-----
function Digits_to_letters (iVec)
implicit none
 	integer		:: i, iVec(N_length)
 	character	:: Digits_to_letters(N_length)

	do i = 1, N_length
		Digits_to_letters(i) = char(iVec(i))
	enddo	

return
end function Digits_to_letters

END PROGRAM Holographic_exemplar_model
